from collections import defaultdict
import itertools
from typing import Dict, Sequence, Tuple
import warnings

from opa_tools.utils import process_dct
from .utils import ALLOWED_KINDS


def flatten(*, type_changed_keys=None, **dcts):
    dct = {}
    for kind, request in dcts.items():
        if kind not in ALLOWED_KINDS:
            warnings.warn(
                f"Adding prefix not listed as allowed kind: '{kind}'."
            )
        dct.update(
            process_dct(
                request,
                prefix=f"{kind}_",
                type_changed_keys=type_changed_keys
            )
        )
    return dct


def unflatten(flatrequest, kinds):
    if kinds is None:
        kinds = ALLOWED_KINDS
    return tuple({
        key.removeprefix(f"{kind}_"): value
        for key, value in flatrequest.items()
        if key.startswith(f"{kind}_")
    } for kind in kinds)


def product(generating_request):
    requests = []
    generating_request_items = list(generating_request.items())
    req_keys, req_values = zip(*generating_request_items)
    for value_comb in itertools.product(*req_values):
        requests.append(dict(zip(req_keys, value_comb)))
    return requests


def wrap(requests: Sequence[Dict]):
    req_dcts = defaultdict(set)
    type_changed_keys = set()
    for flattened_request in requests:
        for key, value in flattened_request.items():
            req_dcts[key].add(value)
    req_dcts = {key: list(values) for key, values in req_dcts.items()}
    if type_changed_keys:
        warnings.warn(
            "Please, review and correct the types for the "
            "following keys in the wrapped request:"
            "\n- " + "\n- ".join(sorted(type_changed_keys))
        )
    return req_dcts
